// opta.ino
//
//	Reads analog inputs from OPTA module
// and writes data to the relays.
//
// Processes the following commands over the Ethernet
// connection:
//
//	ain <nl>		returns "ain xxxx ... xxxx"
//					where "xxxx ... xxxx" represents
//					the eight analog input values
//					(in hexadecimal form). Returns
//					one of the last nine raw readings
//					for each channel from the device.
//
//	ave <nl>		returns "ave xxxx ... xxxx"
//					where "xxxx ... xxxx" represents
//					the eight analog input values
//					(in hexadecimal form). Returns
//					the "mean of medians" for each
// 					channel (median computed across
//					the last nine readings and the
//					mean is computed across the last
//					eight median values).
//
//	dout x x x x	Writes the four values (0 or 1)
//					to the OPTA relays (D0 through D3).
//					Also sets the status LEDs to reflect
//					the state of each relay.


#include <Ethernet.h>

// If DEBUG is set to 1, then this code will
// write debugging information to the serial (USB)
// port.

#define DEBUG 1

// Useful macros:

#define ever ;;
#define breakif( __x__ ) if( __x__ ) break


// Analog Input storage locations:

#define numAIO 8
#define numReadings 9
#define numAVE 8

unsigned short Ain[numAIO][numReadings];
unsigned short Medians[numAIO][numAVE];
unsigned short Mean[numAIO];


// Ethernet stuff.
// Note: mac address comes from the QR code on the bottom of the OPTA.
// IP Address should be in the range 192.168.2.100 to 192.168.2.109
// for analog inputs, or 192.168.2.*, where the LO byte is:
//
//  .30 to .36 for percent power devices (e.g., GA NP)
//  .40 to .46 for linear power devices  (e.g., GA NMP)
//	.50 to .56 for pulsing %pwr devices  (e.g., GA NPP)
//	.60 to .66 for log channel devices	 (e.g., GA NLW)
//	.70 to .76 for fuel temperature devices (e.g, GA NFT)
// 

IPAddress		ip(192,168,2,100);

#define port 45232

byte			mac[]		= {0xA8, 0x61, 0x0A, 0x50, 0x45, 0x97 };
byte 			gateway[] 	= { 192, 168, 2, 1 };
byte 			subnet[] 	= { 255, 255, 255, 0 };

EthernetServer	server 		= EthernetServer( port );
EthernetClient 	client;

// cmdBuf is used to receive commands from client:

#define MSG_BUFF_LEN 256

#define maxResponse (MSG_BUFF_LEN)
int		cmdIndex = 0;
char	cmdBuf[MSG_BUFF_LEN];
char	response[maxResponse];



// Find the median element of a nine-element unsigned short array.
//
//  This Quickselect routine is based on the algorithm described in
//  "Numerical recipes in C", Second Edition,
//  Cambridge University Press, 1992, Section 8.5, ISBN 0-521-43108-5
//  This code by Nicolas Devillard - 1998. Public domain.
//
//	Code was modified to use macros (straight-line code) for arrays with
//	nine elements (an optimization).


#define ELEM_SORT(a,b) { if((a)>(b)) ELEM_SWAP( (a), (b) ); }
#define ELEM_SWAP(a,b) { register int16_t t=(a);(a)=(b);(b)=t; }

unsigned short quick_select(unsigned short array[]) 
{
    int low;
	int	high;
    int median;
    int middle;
	int ll;
	int hh;
	
	unsigned short arr[9];	// Make temporary copy here because we will modify arr
		
	arr[0] = array[0];
	arr[1] = array[1];
	arr[2] = array[2];
	arr[3] = array[3];
	arr[4] = array[4];
	arr[5] = array[5];
	arr[6] = array[6];
	arr[7] = array[7];
	arr[8] = array[8];
	
    ELEM_SORT( arr[1], arr[2] ); 
	ELEM_SORT( arr[4], arr[5] ); 
	ELEM_SORT( arr[7], arr[8] );
    ELEM_SORT( arr[0], arr[1] ); 
	ELEM_SORT( arr[3], arr[4] ); 
	ELEM_SORT( arr[6], arr[7] );
    ELEM_SORT( arr[1], arr[2] ); 
	ELEM_SORT( arr[4], arr[5] ); 
	ELEM_SORT( arr[7], arr[8] );
    ELEM_SORT( arr[0], arr[3] ); 
	ELEM_SORT( arr[5], arr[8] ); 
	ELEM_SORT( arr[4], arr[7] );
    ELEM_SORT( arr[3], arr[6] ); 
	ELEM_SORT( arr[1], arr[4] ); 
	ELEM_SORT( arr[2], arr[5] );
    ELEM_SORT( arr[4], arr[7] ); 
	ELEM_SORT( arr[4], arr[2] ); 
	ELEM_SORT( arr[6], arr[4] );
    ELEM_SORT( arr[4], arr[2] ); 
	return( arr[4]) ;
	
} // quick_select

#undef ELEM_SWAP
#undef ELEM_SORT


// -------------------------------------------------------------------------
// FUNCTION: xtoi
//
// DESCRIPTION: Converts given hex string to int
//
// PARAMETERS: input string representing a hex value
//
// RETURN VALUE: int version of the given string.
//
// SIDE EFFECTS: none
// -------------------------------------------------------------------------

static unsigned long xtoi( char* s, int &error )
{
	unsigned long result	= 0;
	int	i					= -1;
	int len 				= (int)strlen( s );

	error = false;
	if( len==0 || len > 8 )
	{
		error = true;
		return 0;
	}
	while( isxdigit( s[++i] ))
	{
		result =
				(result << 4)
			+	(
						(s[i] >= 'A')
					?	((s[i] & 0xf) + 9)	// A-F or a-f (A=0x41, a=0x61)
					:	s[i] & 0xf			// 0-9 (0=0x30)
				);
	}
	error = s[i] != 0 && !isspace( s[i] );
	return result;

} // end xtoi


// -------------------------------------------------------------------------
// FUNCTION: skipSpaces
//
// DESCRIPTION:
//     Given an asciiz string, advances to the first non-space (or end of
//     string)
//
// PARAMETERS:
//		msg: asciiz string
//		ndx: index pointer into the 'msg' string
//
// RETURN VALUE:
//		index into string where next non-space is found.
//
// SIDE EFFECTS: none
//
// --------------------------------------------------------------------------

static int skipSpaces( char *msg, int ndx )
{
	for(ever)
	{
		char ch = msg[ndx];

		breakif( ch==0 || (ch!=' ' && ch!='\t' && ch!=':') ); //hit non-space or end of message
		++ndx;		//increment past spaces
	}
	return ndx;

} // end skipSpaces


// -------------------------------------------------------------------------
// FUNCTION: nextWord
//
// DESCRIPTION:
//     Given an asciiz string 'msg', returns space delimited asciiz 'word'.
//     Skips leading spaces. Leaves 'ndx' at the first space following the
//     'word' (or at the terminating null), or at 'maxLen' if max word size
//     was reached.
//     Return value is 'word' and an updated position for 'ndx'
//
// PARAMETERS:
//     msg: given message, an asciiz string.
//     ndx: index pointing to the current position in 'msg'
//     word: returned word extracted from 'msg'
//     maxlen: maximum length allowed for 'word'
//
// RETURN VALUE: true if end of 'msg' is reached.
//
// SIDE EFFECTS:
//		Modifies integer object pointed at by ndx ref parm.
//		Stores new word in array pointed at by "word".
//
// -------------------------------------------------------------------------

static int nextWord( char *msg, int &ndx, char *word, int maxLen )
{
	int		i	= 0; //index into the parsed word
	char	ch	= 0;

	ndx = skipSpaces( msg, ndx ); //eliminate any leading spaces
	for(ever)
	{
		//get first word of message

		breakif( i >= (maxLen-1) );	//exceeded maximum word length
		ch = msg[ndx];

		// See if we've hit delimiter or end of incoming message

		breakif( ch==0 || ch=='\n' || ch==' ' || ch=='\t' || ch==':' || ch=='\r' );

		// Build up the word:

		word[i++] = msg[ndx++];
	}
	word[i] = 0;	//terminate the word
	return ch==0;	//return true on end of message

} // nextWord


/////////////////////////////////////////////////////////////////////////
//
//  FUNCTION: execCmd
//
//  DESCRIPTION: This is called when a complete Ethernet message arrives.
//
//  PARAMETERS:
//      msg: incoming message. ASCIIZ string.
//      response: reponse message (if any) is copied into this buffer
//      maxRespLen: the max allocated buffer size for response message.
//
//  RETURN VALUE: length of 'response' message.


#define strieq( s1, s2 ) (strcasecmp((s1), (s2)) == 0 )	// Case insensitive string compare

void execCmd( char * msg, char *response, int maxRespLen  )
{
	int				badParam		= false;        	//Flags illegal parameters
	int				error			= false;        	//Flags illegal parameters
	int				paramCnt		= 0;        		//no. of parameters found
	int				ndx				= 0;        		//index into the incoming msg

	unsigned 		params[MSG_BUFF_LEN];   			//message parameters (subsequent words)
	char			msgCopy[MSG_BUFF_LEN];  			//command (first word in msg)
	char			cmd[MSG_BUFF_LEN];  				//command (first word in msg)
	char			parmBuf[MSG_BUFF_LEN]; 				//pulse buffer for extracting parameters


	// Begin by clearing the response string (if there is at least one byte available
	// in the array):

	if( maxRespLen > 0 )
	{
		*response = 0;
	}

	// Remove trailing spaces from the command string:

	int i = strlen( msg );
	while( --i > 0 && isspace( msg[i] ));
	msg[i + 1] = 0;

	// Make a copy of the incoming message so we can print its contents
	// if there is an error:

	strncpy( msgCopy, msg, MSG_BUFF_LEN-1 );
	msgCopy[MSG_BUFF_LEN-1] = 0;

	// Grab the first "word" in the string. This will be the command ("ain", "dout",
	// or "ave").

	nextWord( msg, ndx, cmd, sizeof cmd );				// get command

	// The remaining items in the string must be parameters. Extract them and
	// save their values in the params array. Note that parameters are always
	// hexadecimal numbers.

	paramCnt	= 0;									// extract the parameters (if any)
	badParam	= 0;									// Assume args are good
	while( paramCnt < MSG_BUFF_LEN )
	{
		nextWord( msg, ndx, parmBuf, sizeof parmBuf );
		breakif(parmBuf[0]==0);								// All done
		params[paramCnt++] = xtoi(parmBuf, error); 			//convert hex string to int
		badParam |= error;

	} // endwhile


	// Handle dout command (write data to OPTA relays):

	if( !badParam && strieq( cmd, "dout" ))
	{
		if( paramCnt == 4 )
		{
			snprintf
			(
				response,
				maxRespLen,
				"dout %03x %03x %03x %03x\n",
				params[0],
				params[1],
				params[2],
				params[3]
			);
			
			// Write the data to the OPTA relays:
			
			digitalWrite( D0, !!params[0] ); 
			digitalWrite( D1, !!params[1] ); 
			digitalWrite( D2, !!params[2] ); 
			digitalWrite( D3, !!params[3] );
			
			// Mirror relay state to the LEDs:
			
			digitalWrite( LED_D0, !!params[0] ); 
			digitalWrite( LED_D1, !!params[1] ); 
			digitalWrite( LED_D2, !!params[2] ); 
			digitalWrite( LED_D3, !!params[3] );
		}
		else
		{
			snprintf
			(
				response,
				maxRespLen,
				"Error: DOUT must have 4 arguments\n"
			);
		}


	} // if dout
	else if( !badParam && strieq( cmd, "ain" ))
	{
		if( paramCnt == 0 )
		{
			// Return raw analog input values.
			// Note that it doesn't better which of
			// the nine entries in Ain this code
			// returns; the sample rate is high enough
			// that any entry will suffice.
			
			snprintf
			(
				response,
				maxRespLen,
				"ain %03x %03x %03x %03x %03x %03x %03x %03x\n",
				Ain[0][0],
				Ain[1][0],
				Ain[2][0],
				Ain[3][0],
				Ain[4][0],
				Ain[5][0],
				Ain[6][0],
				Ain[7][0]
			);
		}
		else
		{
			snprintf
			(
				response,
				maxRespLen,
				"Error: Ain doesn't support any arguments\n"
			);
		}


	} // if ain
	else if( !badParam && strieq( cmd, "ave" ))
	{
		if( paramCnt == 0 )
		{
			// Return filtered analog input values.
			
			snprintf
			(
				response,
				maxRespLen,
				"ave %03x %03x %03x %03x %03x %03x %03x %03x\n",
				Mean[0],
				Mean[1],
				Mean[2],
				Mean[3],
				Mean[4],
				Mean[5],
				Mean[6],
				Mean[7]
			);
		}
		else
		{
			snprintf
			(
				response,
				maxRespLen,
				"Error: AVE doesn't support any arguments\n"
			);
		}


	} // if ain
	else if( !badParam && strlen( cmd ) == 0 )
	{
		// Ignore empty command strings

		response[0] = 0;
	}

	////////////////////////////////////////////////////
	//
	// If none of the above, we have an error.

	else
	{
		snprintf( response, maxRespLen, "Error: unrecognized command in execCmd: %s\n", msg);
	}
	return;

} // end execCmd


/////////////////////////////////////////////////////////////////////
//
// processAIN
//
//	Passed a value read from analog input.
// Will update the Ain, Medians, and Mean arrays.

void processAIN( int which, int value, int ainIndex, int medianIndex )
{
	// Remove the oldest entry from Ain and insert the newest entry:
	
	Ain[which][ainIndex] = (unsigned short) (value & 0xffff);
	
	// Compute the median of the last nine input values and
	// save in the Medians array:
	
	Medians[which][medianIndex] = quick_select( Ain[which] );
	
	// Compute the mean of the last eight median readings:
	
	unsigned u = 0;
	for( int i=0; i < numAVE; ++i )
	{
		u += Medians[which][i];
	}
	Mean[which] = (unsigned short) (u / numAVE);
}

/////////////////////////////////////////////////////////////////////
//
// setup
//
// Program initialization:

void setup() 
{
	int buttonState;
	
	pinMode(LED_BUILTIN, OUTPUT);
	pinMode(LED_D0, OUTPUT);
  	pinMode(LED_D1, OUTPUT);
  	pinMode(LED_D2, OUTPUT);
  	pinMode(LED_D3, OUTPUT);
	digitalWrite( LED_D0, HIGH );
	digitalWrite( LED_D1, HIGH );
	digitalWrite( LED_D2, HIGH );
	digitalWrite( LED_D3, HIGH );

	Serial.begin( 9600 );
	while( !Serial );
	
	// Initialize I/O on the OPTA:

	// LEDs:

	digitalWrite( LED_D0, LOW );
	digitalWrite( LED_D1, LOW );
	digitalWrite( LED_D2, LOW );
	digitalWrite( LED_D3, LOW );
	
	
	// User Button:
	
    pinMode(BTN_USER, INPUT);
	
	// Relays:
	
	pinMode(D0, OUTPUT);
  	pinMode(D1, OUTPUT);
  	pinMode(D2, OUTPUT);
  	pinMode(D3, OUTPUT);

	// Analog input resolution (A0-A7):
	
	analogReadResolution(16);
	
	// Ethernet initialization:
	
	Ethernet.begin( mac, ip, gateway, subnet );

	// Check for Ethernet hardware present

	if (Ethernet.hardwareStatus() == EthernetNoHardware)
	{
		#if DEBUG
		
			Serial.println("Ethernet shield was not found.  Sorry, can't run without hardware. :(");
			
		#endif
		while (true)
		{
		 	// Indicate status error on LEDs:
			
			digitalWrite( LED_D0, LOW );
			digitalWrite( LED_D1, LOW );
			digitalWrite( LED_D2, LOW );
			digitalWrite( LED_D3, LOW );
		  	delay( 100 );
			digitalWrite( LED_D0, HIGH );
			digitalWrite( LED_D1, HIGH );
			digitalWrite( LED_D2, HIGH );
			digitalWrite( LED_D3, HIGH );
		  	delay( 100 );
		}
	}
	
	// Wait for the IP address to be set:
	
	while( Ethernet.localIP() != ip );

	// Start listening for a connection:
	
	server.begin();
	
	#if DEBUG
	
		Serial.println( "Initialization complete" );
		Serial.print( "IP Address = " );
		Serial.print( Ethernet.localIP() );
		Serial.print( " port(" );
		Serial.print( port );
		Serial.println( ")" );
		
	#endif
}


// the loop function runs over and over again forever

void loop() 
{
	static int	Aindex = 0;
	static int	Mindex = 0;
	static int	ledState = 0;
	static int	lastTime = 0;
	static int  blinkCnt = 1500;
	int 		sawEOLN = 0;
		
	int time = millis();

	if( time > lastTime + blinkCnt )
	{
		lastTime = time;
		ledState = !ledState;
	}
	digitalWrite( LED_BUILTIN, ledState );
		 
	
	processAIN( 0, analogRead(A0), Aindex, Mindex );
	processAIN( 1, analogRead(A1), Aindex, Mindex );
	processAIN( 2, analogRead(A2), Aindex, Mindex );
	processAIN( 3, analogRead(A3), Aindex, Mindex );
	processAIN( 4, analogRead(A4), Aindex, Mindex );
	processAIN( 5, analogRead(A5), Aindex, Mindex );
	processAIN( 6, analogRead(A6), Aindex, Mindex );
	processAIN( 7, analogRead(A7), Aindex, Mindex );
	Aindex = (Aindex + 1) % numReadings;
	Mindex = (Mindex + 1) % numAVE;	
	
	// Transmit on Ethernet port

	EthernetClient newClient = server.available();
	if( newClient )
	{
		if( !client )
		{
			blinkCnt = 250;
			client = newClient;
			#if DEBUG
			
				Serial.println( "Client connected" );
				
			#endif
		}
		else
		{
			#if DEBUG
			
				Serial.println( "New client connection refused" );
				
			#endif
			newClient.stop();
		} // endif
	}// endif
	if( client )
	{
		if( client.available() > 0 )
		{
			int ch = client.read();
			while( ch != -1 )
			{
				cmdBuf[cmdIndex] = ch;
				if( ch == '\n' )
				{
					cmdBuf[ cmdIndex ] = 0;
					cmdIndex = 0;
					#if DEBUG
					
						Serial.print( "Received: '" );
						Serial.print( cmdBuf );
						Serial.println( "'" );
					
					#endif					
					execCmd( cmdBuf, response, maxResponse );
					client.write( response, strlen( response ) );
					break;
				}
				else if( ch != '\r' && cmdIndex < sizeof( cmdBuf ))
				{
					++cmdIndex;
				}
				ch = client.read();
			} // endwhile			
		} // endif client.available() > 0
	} // endif
	
	// Stop any clients of server which disconnect

	if (client && !client.connected())
	{
		blinkCnt = 1500;
		#if DEBUG
		
			Serial.println( "Disconnecting from client" );
			
		#endif
		client.stop();
	}
	
	
}



